// maptest.cpp : exercises a map of words

#include "stdafx.h"
#include <iostream>
#include <fstream>
#include <map>
#include <string>
using namespace std;

const int WORDSIZE=32;
const int BUFFSIZE = 1024;
const char *DELIM = " \t\n;,.?";  // marks ends of words

// this class holds each word we find. There will be one of these
// in the map for each unique word in the text.
// Each Word object will keep a count of how many times it appears in the
// text.
class Word 
{
public:
    Word() {count = 0;}
    Word (string str) { data = str; count = 1; }
    string get_word() { return(data); }
    void bump_count() { ++count;}
    int get_count() { return(count);}
    friend ostream & operator<<(ostream & os, Word & w) 
    {
        os << w.data << " --> " << w.count << endl;
        return os;
    }
private:
        string data; // holds the word
        int count; // holds the count of times it appears
};
// create a shorthand name for the actual map object
// uses the word string as a key to find the Word object
// Note that a pointer to the Word is stored not the actual object
typedef map<string, Word *> MYMAP;

// this holds the map object and routines to use it
class My_map {
    public:
        void insert(Word *);    // add a word to the map
        int size() { return count_table.size();} // how many words
        void print(ostream &); // display the contents of the map
        void parse_line(char * ); // break a line of text into words
        int total_words();  // count the total words in the document
        string get_first();  // alphabetically least word
        string get_last();   // alphabeticall greatest word.
    // the operator[] exists for the map object so we have to
    // provide one to so we can retrieve the counts using the
    // word as in index. Note that this returns an integer
    // and can only be used to get the values, not change them.
    // An example of its use is
    // cout << table["hello"];
    // the table["hello"] part becomes
    // table.operator[]("hello")
    int operator[](string s) 
    { 
//        Word * w = count_table[s];
//        return w->get_count();
        // this could be written as
        return (*count_table[s]).get_count();
        // because the count_table[s] part returns a Word pointer.
        // after we put the * in front of it, it is a Word object.
        // So we can call the get_count() method.
        }
    private:
        MYMAP count_table;
};

// wrapper around map insert
void
My_map::insert(Word *w)
{
    string str = w->get_word();  // get the text of the word
    // search the map to find out if the word is already in there
    MYMAP::iterator it = count_table.find(str);

    if(it == count_table.end()) { // word not found
        //  the map holds a collection of things called 'value_type' and
        //  the map<string, Word *>::value_type(str,w) part creates
        //  a new one with our key (str) and the new Word object (w)
        count_table.insert(MYMAP::value_type(str,w));
    }
    else {
        it->second->bump_count(); // already there, increase count
        }
} // insert

// split the line into words using the characters
// in DELIM as end of word markers
// when strtok is called with a NULL pointer
// it continues on the same line where it left off
// it returns NULL when there is no more text
void
My_map::parse_line(char * line)
{
    char * str;
    Word * w;
    str = strtok(line,DELIM);  // find the first word in line
    while(str != NULL) {
        // make lower case
        for(int i=0; str[i] ; i++) str[i] = tolower(str[i]);
        if((w = new Word(string(str))) == NULL) throw "no mem";
        insert(w);
        str = strtok(NULL,DELIM); // find the next word in line
    } // while
} // parse_line

// the argument is the output stream.
// Could be a file, could be cout
void
My_map::print(ostream & out)
{
    MYMAP::iterator it;
    for(it=count_table.begin(); it != count_table.end(); it++)
        out << *(it->second);
} // print

// return the total count of words by summing the counts
// from each word in the map
int
My_map::total_words()
{
    MYMAP::iterator it;
    int count = 0;
    for(it=count_table.begin(); it != count_table.end(); it++)
        count += it->second->get_count();
    return(count);
} // total_words

string
My_map::get_first() {
    MYMAP::iterator it = count_table.begin();
    return(it->second->get_word());
} // get_first

// a reverse iterator starts at the end of the collection and
// works towards the front.
// So rbegin() returns an iterator pointing to the last object in the collection
// rather than past the last one like end() does.
string
My_map::get_last() {
    MYMAP::reverse_iterator it = count_table.rbegin();
    return(it->second->get_word());
} // get_last

int 
main(int argc, char* argv[])
{
     ifstream infile;
     ofstream outfile;
     char word[2*WORDSIZE];
     char line[BUFFSIZE];
     My_map table;

     infile.open("mapdata.txt",ios::in);
     if(!infile) {
         cerr << "open of input file failed\n";
         exit(1);
     }
     outfile.open("mapout.txt",ios::out);
     if(!outfile) {
         cerr << "open of output file failed\n";
         exit(1);
     }
     
     // insert some data
     try {
          while(infile.getline(line,BUFFSIZE,'\n')) {
              table.parse_line(line);
          } // while
     } catch (char * str) {
          cout << "insert error " << str << endl;
          exit(1);
       }

     table.print(outfile); // save to file
     table.print(cout);    // show on display
     cout << "there were " << table.size() << " unique words\n";
     cout << "there were " << table.total_words() << " total words\n";
     cout << "the smallest word is " << table.get_first() << endl;
     cout << "the biggest word is " << table.get_last() << endl;
     cout << "the word a appeared " << table["a"] << " times" << endl;
     return 0;
} // main
